import httpx
import requests
import os


def url_to_filename(url, is_ext=False):
    filename = os.path.basename(url)
    if not is_ext:
        filename = filename[:filename.find('.')]
    return filename


def requests_get_bytes(url):
    r = requests.get(url)
    if r.status_code == 200:
        bytes1 = r.content
        if len(bytes1) > 180:  # size larger than 180 bytes we assume the file is not empty
            return bytes1
        else:
            return b''


def requests_get_ram_cache(f, url, target_directory):
    return f(requests_get_bytes(url))


async def async_post(url, dic):
    async with httpx.AsyncClient() as c:
        return (await c.post(url, data=dic)).json()


async def async_get(url):
    async with httpx.AsyncClient() as c:
        return (await c.get(url)).json()



import hashlib
import os
from os.path import exists

import requests

from cameo_claw import it_distinct, it_group, it_filter, server, it_filter_group, it_streaming_pl, it_download, \
    it_filter_group_mean_count, it_filter_group_append
import time
from cameo_claw.log import log
import cameo_claw as cc


def get_lst_url_taipei_100():
    lst_device_id = [
        "10180507985",
        "10183804818",
        "10180935138",
        "10182362729",
        "10186915768",
        "10182765572",
        "10183643733",
        "10187817545",
        "10180711468",
        "10181888891",
        "10184237251",
        "10200376949",
        "10187714974",
        "10188959173",
        "10181133624",
        "10189875513",
        "10184780057",
        "10187536226",
        "10188542674",
        "10187616860",
        "10190212306",
        "10200485379",
        "10184404771",
        "10188107333",
        "10186524358",
        "10187942042",
        "10183163884",
        "10180654119",
        "10185092420",
        "10187452115",
        "10190000864",
        "10190479538",
        "10185579073",
        "10181983795",
        "10188844366",
        "10189360662",
        "10182995926",
        "10181055695",
        "10189089883",
        "10181464107",
        "10184979038",
        "10181258942",
        "10184871059",
        "10185137651",
        "10182074994",
        "10190374312",
        "10186752944",
        "10189461434",
        "10184589018",
        "10181395211",
        "10186372622",
        "10186223376",
        "10182118699",
        "10190513901",
        "10185716972",
        "10189779191",
        "10188219952",
        "10185390063",
        "10181621345",
        "10182567005",
        "10186123517",
        "10188652949",
        "10184615499",
        "10189289540",
        "10182827296",
        "10185414614",
        "10184032107",
        "10185295344",
        "10182603505",
        "10186080368",
        "10187340888",
        "10186497242",
        "10187268812",
        "10183739942",
        "10188795449",
        "10183269165",
        "10183435474",
        "10180884769",
        "10188098571",
        "10183378041",
        "10185861505",
        "10183938343",
        "10185920135",
        "10181700710",
        "10188450871",
        "10186643177",
        "10184128532",
        "10190122978",
        "10188364226",
        "10184373796",
        "10186871514",
        "10181510720",
        "10183091769",
        "10187195615",
        "10182493773",
        "10187012610",
        "10182273531",
        "10183554350",
        "10185613026",
        "10189170767",
    ]
    return get_lst_url_by_date(lst_device_id)


def get_lst_url_50():
    lst_device_id = [
        "11144353041",
        "11135655524",
        "11146220817",
        "11134915729",
        "11144949700",
        "11145345031",
        "11144535597",
        "11150348864",
        "11152332479",
        "11137445267",
        "11142706581",
        "11139427726",
        "11132757143",
        "11143315941",
        "11145428733",
        "11138263064",
        "11149040829",
        "11135424517",
        "11147350591",
        "11132265371",
        "11136533184",
        "11141843507",
        "11136480034",
        "11149117865",
        "11148832575",
        "11133512028",
        "11143988456",
        "11143119216",
        "11134364858",
        "11143240238",
        "11132477607",
        "11146482455",
        "11131367272",
        "11151053715",
        "11138740618",
        "11143570743",
        "11132083287",
        "11135577162",
        "11144277295",
        "11134472620",
        "11151283365",
        "11147041406",
        "11131824250",
        "11136322895",
        "11149357462",
        "11131209112",
        "11139221095",
        "11148103547",
        "11145230386",
        "11132685693",
    ]
    return get_lst_url_by_date(lst_device_id)


def get_lst_url_by_date(lst_device_id):
    lst_url = []
    for device_id in lst_device_id:
        # for i in range(1, 32):
        for i in range(1, 2):
            day = f'2022-03-{i:02}'
            url = f'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.industry.rawdata.material/device_{device_id}/device_{device_id}_daily_{day}.csv.gz'
            lst_url.append(url)
    return lst_url


# def test_filter():
#     lst_select_column = ['createTime', 'deviceId', 'lat', 'localTime', 'lon', 'sensorId', 'value']
#     lst_distinct_column = ['deviceId', 'localTime', 'sensorId']
#     filter_column = 'sensorId'
#     filter_value = 'pm2_5'
#     target_directory = './data/parquet/'
#     multiprocessing_download_to_parquet(
#         get_lst_url(),
#         lst_select_column,
#         lst_distinct_column,
#         filter_column,
#         filter_value,
#         target_directory)

def test_it_download():
    target_directory = './data/topic_download/'
    int_total = len(get_lst_url_taipei_100())
    for int_progress, done_url in it_download(get_lst_url_taipei_100(), target_directory):
        # print_progress(done_url, int_progress, int_total)
        pass


def test_it_distinct():
    target_directory = './data/topic_distinct/'
    int_total, lst_distinct_column = total_distinct()
    for int_progress, done_url in it_distinct(get_lst_url_taipei_100(), target_directory, lst_distinct_column):
        # print_progress(done_url, int_progress, int_total)
        pass


def total_distinct():
    int_total = len(get_lst_url_taipei_100())
    lst_distinct_column = ['deviceId', 'localTime', 'sensorId']
    return int_total, lst_distinct_column


def print_progress(done_url, int_progress, int_total):
    print(f'int_progress/int_total = {int_progress}/{int_total}')
    print(f'done url: {done_url}')


def test_it_group():
    target_directory = './data/topic_group/'
    int_total, lst_distinct_column = total_distinct()
    lst_group_by_column = ['deviceId', 'sensorId']
    for int_progress, done_url in it_group(get_lst_url_taipei_100(), target_directory, lst_distinct_column,
                                           lst_group_by_column):
        print_progress(done_url, int_progress, int_total)


def test_it_filter():
    target_directory = './data/topic_filter/'
    int_total, lst_distinct_column = total_distinct()
    lst_column_match = [['sensorId', 'pm2_5'], ]
    sort_column = 'localTime'
    for int_progress, done_url in it_filter(
            get_lst_url_taipei_100(), target_directory, lst_distinct_column,
            lst_column_match, sort_column):
        pass
        # print_progress(done_url, int_progress, int_total)


def test_server():
    server()


def test_client():
    print('test.py test_client!')
    d = {'lst_url': get_lst_url_taipei_100(),
         'target_directory': './data/topic_filter/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime'}
    print(d)
    r = requests.post('http://localhost:20409/api/streaming_post/', json=d)
    print(f'http://localhost:20409/api/streaming_get/?md5_hex={r.text[1:-1]}')
    with open('all.csv.gz', 'wb') as f:
        f.write(r.content)


def test_it_filter_group():
    target_directory = './data/topic_filter_group/'
    int_total, lst_distinct_column = total_distinct()
    lst_column_match = [['sensorId', 'pm2_5'], ['sensorId', 'voc']]
    sort_column = 'localTime'
    lst_group_by_column = ['deviceId']
    for int_progress, done_url in it_filter_group(
            get_lst_url_taipei_100(), target_directory, lst_distinct_column,
            lst_column_match, sort_column, lst_group_by_column):
        pass


# def test_it_filter_group_append():
#     lst_url = get_lst_url_taipei_100()
#     target_directory = './data/topic_filter_group_append/'
#     int_total, lst_distinct_column = total_distinct()
#     lst_column_match = [['sensorId', 'pm2_5'], ['sensorId', 'voc']]
#     sort_column = 'localTime'
#     lst_group_by_column = ['deviceId']
#     append_column = 'deviceId'
#     column_header = 'createTime,deviceId,lat,localTime,lon,sensorId,value\n'
#
#     target_directory = cache_directory(lst_url, target_directory, lst_distinct_column, lst_column_match,
#                                        sort_column, lst_group_by_column, append_column, column_header)
#
#     if exists(target_directory):
#         print('cameo_claw.py, it_filter_group_append, target_directory already existed.', target_directory)
#         return False
#
#     for int_progress, done_url in it_filter_group_append(
#             lst_url, target_directory, lst_distinct_column,
#             lst_column_match, sort_column, lst_group_by_column, append_column, column_header):
#         pass
#     return True


def test_it_filter_group_mean_count():
    target_directory = './data/topic_filter_group_mean_count/'
    int_total, lst_distinct_column = total_distinct()
    lst_column_match = [['sensorId', 'pm2_5'], ['sensorId', 'voc']]
    sort_column = 'localTime'
    lst_group_by_column = ['deviceId', 'sensorId']
    time_column = 'localTime'
    value_column = 'value'
    for int_progress, done_url in it_filter_group_mean_count(
            get_lst_url_taipei_100(), target_directory, lst_distinct_column,
            lst_column_match, sort_column, lst_group_by_column, time_column, value_column):
        pass


# def pipe_distinct(url, d):
#     df = cc.read_csv(url)
#     df = cc.distinct(df, d['lst_distinct_column'])
#     cc.to_csv(df, d['target_directory'] + cc.url_to_filename(url))


def pipe_epa_download(url, d):
    df = cc.read_csv(url)
    if df:
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        g = cc.groupby(df, d['lst_group_by_column'])
        for df in g:
            cc.append_csv(df, d)
    return url


# def test_epa_download():
#     d = {'lst_url': get_lst_url_taipei_100(),
#          'target_directory': './data/topic_epa_download/',
#          'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
#          'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
#          'sort_column': 'localTime',
#          'lst_group_by_column': ['deviceId'],
#          'append_column': 'deviceId',
#          'column_header': 'createTime,deviceId,lat,localTime,lon,sensorId,value\n'}
#     d['target_directory'] = cc.cache_directory(d)
#     zip_path = cc.zip_path(d['target_directory'])
#     if not exists(zip_path):
#         for int_progress, result in cc.it(pipe_epa_download, d['lst_url'], d):
#             pass
#         cc.zip(d['target_directory'] + '*.csv.gz', zip_path)
#     return zip_path


def test_epa_download():
    d = {'lst_url': get_lst_url_taipei_100(),
         'target_directory': './data/topic_epa_download/',
         'lst_distinct_column': ['deviceId', 'localTime', 'sensorId'],
         'lst_column_match': [['sensorId', 'pm2_5'], ['sensorId', 'voc']],
         'sort_column': 'localTime',
         'lst_group_by_column': ['deviceId'],
         'append_column': 'deviceId',
         'column_header': 'createTime,deviceId,lat,localTime,lon,sensorId,value\n'}
    if cc.has_zip_cache(d):
        zip_cache_path = cc.zip_cache_path(d)
        print('zip_cache_path existed:', zip_cache_path)
        return zip_cache_path
    for int_progress, result in cc.pipe(pipe_epa_download, d['lst_url'], d):
        pass
    return cc.zip_cache(d)


if __name__ == '__main__':
    # print('sh/go.sh server')
    # print('sh/go.sh client')
    # print('sh/go.sh filter')
    # if sys.argv[1] == 'server':
    #     test_server()
    # if sys.argv[1] == 'client':
    #     test_client()

    # t = time.time()
    # test_it_filter()
    # log(f'test_it_filter:{time.time() - t}')

    # t = time.time()
    # test_it_download()
    # log(f'test_it_download:{time.time() - t}')

    # t = time.time()
    # test_it_distinct()
    # log(f'test_it_distinct:{time.time() - t}')

    # t = time.time()
    # test_it_group()
    # log(f'test_it_group:{time.time() - t}')

    # t = time.time()
    # test_it_filter_group()
    # log(f'test_it_group:{time.time() - t}')

    # t = time.time()
    # test_it_filter_group_append()
    # log(f'test_it_group:{time.time() - t}')

    # t = time.time()
    # test_it_filter_group_mean_count()
    # log(f'test_it_group:{time.time() - t}')

    # t = time.time()
    # test_epa_download()
    # log(f'test_group_append:{time.time() - t}')

    t = time.time()
    test_epa_download()
    log(f'test_group_append:{time.time() - t}')

# sh/go.sh



import gzip
import os
from io import BytesIO
from zipfile import ZipFile

from filelock import FileLock
from os.path import exists

from cameo_claw.cameo_claw_reuse import bytesio_to_gzip, requests_get_write, \
    pl_read_csv_distinct, groupby_write_gzip, it_f, condition_filter_sort, distinct_filter_sort, \
    bytesio_filename, write_gzip, df_write_gzip, condition, df_write_csv
import polars as pl

from cameo_claw.file import mkdir
from cameo_claw.functional import it_mp_f
from cameo_claw.net import requests_get_bytes
from glob import glob
from functools import lru_cache


def a_download(url, target_directory):
    def write(bytes1):
        filename = os.path.basename(url)
        path = f'{target_directory}{filename}'
        with open(path, 'wb') as f:
            f.write(bytes1)
        return url

    return requests_get_write(write, target_directory, url)


def it_download(lst_url, target_directory):
    return it_mp_f(a_download, [tuple([url, target_directory]) for url in lst_url])


def a_distinct(url, target_directory, lst_distinct_column):
    def write(bytes1):
        df = pl_read_csv_distinct(bytes1, lst_distinct_column)
        return bytesio_to_gzip(df, url, target_directory)

    return requests_get_write(write, target_directory, url)


def it_distinct(lst_url, target_directory, lst_distinct_column):
    return it_mp_f(a_distinct, [tuple([url, target_directory, lst_distinct_column]) for url in lst_url])


def a_group(url, target_directory, lst_distinct_column, lst_group_by_column):
    def write(bytes1):
        df = pl_read_csv_distinct(bytes1, lst_distinct_column)
        return groupby_write_gzip(df, url, lst_group_by_column, target_directory)

    return requests_get_write(write, target_directory, url)


def it_group(lst_url, target_directory, lst_distinct_column, lst_group_by_column):
    return it_mp_f(a_group,
                   [tuple([url, target_directory, lst_distinct_column, lst_group_by_column]) for url in lst_url])


def a_filter(url, target_directory, lst_distinct_column, lst_column_match=[], sort_column=''):
    def write(bytes1):
        df = distinct_filter_sort(bytes1, lst_distinct_column, lst_column_match, sort_column)
        bytesio_to_gzip(df, url, target_directory)
        return url

    return requests_get_write(write, target_directory, url)


def it_filter(lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column):
    return it_f(a_filter, lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column)


def a_chunk_pl_lst(url, target_directory, lst_distinct_column, lst_column_match, sort_column):
    tup_distinct_column = tuple(lst_distinct_column)
    tup_lst_column_match = tuple([tuple(l) for l in lst_column_match])
    return a_chunk_pl_tup(url, target_directory, tup_distinct_column, tup_lst_column_match, sort_column)


@lru_cache(maxsize=5000)
def a_chunk_pl_tup(url, target_directory, tup_distinct_column, tup_column_match, sort_column):
    def write(bytes1):
        df = pl.read_csv(bytes1)
        df = df.distinct(subset=list(tup_distinct_column))
        df = condition_filter_sort(df, tup_column_match, sort_column)
        bytesio_csv = BytesIO()
        df.to_csv_gz(bytesio_csv, has_header=False)
        bytesio_gzip = BytesIO()
        with gzip.open(bytesio_gzip, 'wb') as f:
            f.write(bytesio_csv.getvalue())
        return bytesio_gzip.getvalue()

    return requests_get_write(write, target_directory, url)


def it_chunk_pl(lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column):
    return it_f(a_chunk_pl_lst, lst_url, target_directory, lst_distinct_column, lst_column_match, sort_column)


def a_filter_group(url, target_directory, lst_distinct_column, lst_column_match=[], sort_column='',
                   lst_group_by_column=[]):
    def write(bytes1):
        df = distinct_filter_sort(bytes1, lst_distinct_column, lst_column_match, sort_column)
        return groupby_write_gzip(df, url, lst_group_by_column, target_directory)

    return requests_get_write(write, target_directory, url)


def it_filter_group(lst_url, target_directory, lst_distinct_column,
                    lst_column_match, sort_column, lst_group_by_column):
    return it_mp_f(a_filter_group,
                   [tuple([url, target_directory, lst_distinct_column,
                           lst_column_match, sort_column, lst_group_by_column]) for url in lst_url])


def a_filter_group_append(url, target_directory, lst_distinct_column, lst_column_match,
                          sort_column, lst_group_by_column, append_column, column_header):
    def write(bytes1):
        df = distinct_filter_sort(bytes1, lst_distinct_column, lst_column_match, sort_column)
        g = df.groupby(lst_group_by_column)
        for df in g:
            bytesio = BytesIO()
            df_write_csv(df, bytesio, has_header=False)
            path = f'{target_directory}{append_column}_{str(df.row(0)[df.find_idx_by_name(append_column)])}.csv.gz'
            with FileLock(f'{path}.lock', timeout=60 * 60 * 2):
                if not exists(path):
                    with gzip.open(path, 'wb') as f:
                        f.write(column_header.encode('utf-8'))
                write_gzip(bytesio, path, 'ab')
        return url

    return requests_get_write(write, target_directory, url)


import hashlib


def it_filter_group_append(lst_url, target_directory, lst_distinct_column, lst_column_match,
                           sort_column, lst_group_by_column, append_column, column_header):
    return it_mp_f(a_filter_group_append,
                   [tuple([url, target_directory, lst_distinct_column,
                           lst_column_match, sort_column, lst_group_by_column, append_column, column_header]) for url in
                    lst_url])


def a_filter_group_mean_count(url, target_directory, lst_distinct_column,
                              lst_column_match, sort_column, lst_group_by_column,
                              time_column, value_column):
    def write(bytes1):
        df = distinct_filter_sort(bytes1, lst_distinct_column, lst_column_match, sort_column)
        for df in df.groupby(lst_group_by_column):
            filename_tail = df_write_gzip(df, True, lst_group_by_column, target_directory, url)
            df_mean_daily = (
                df.lazy()
                    .groupby(lst_group_by_column)
                    .agg(
                    [
                        pl.first('localTime'),
                        pl.col(value_column).mean(),
                    ]
                )
            ).collect()
            bytesio, filename = bytesio_filename(df_mean_daily, url)
            path = f'{target_directory}{filename}{filename_tail}_mean_daily.csv.gz'
            write_gzip(bytesio, path)

    return requests_get_write(write, target_directory, url)


def it_filter_group_mean_count(lst_url, target_directory, lst_distinct_column,
                               lst_column_match, sort_column, lst_group_by_column,
                               time_column, value_column):
    return it_mp_f(a_filter_group_mean_count,
                   [tuple([url, target_directory, lst_distinct_column,
                           lst_column_match, sort_column, lst_group_by_column,
                           time_column, value_column]) for url in lst_url])


def md5_hex(d):
    return hashlib.md5(str(d).encode('utf-8')).hexdigest()


# def cache_directory(d):
#     return d['target_directory'] + f'{md5_hex(d)}/'


def zip_cache_path(d):
    hash = md5_hex(d)
    return d['target_directory'] + f'{hash}/{hash}.zip'


def read_csv(url):
    return pl.read_csv(requests_get_bytes(url), infer_schema_length=20000)


def distinct(df, lst_distinct_column):
    return df.distinct(subset=lst_distinct_column)


def filter(df, lst_column_match):
    return df.filter(eval(condition(lst_column_match)))


def sort(df, sort_column):
    return df.sort(sort_column)


def groupby(df, lst_group_by_column):
    return df.groupby(lst_group_by_column)


def append_csv_hash(df, target_directory, append_column, column_header, hash_path):
    mkdir(target_directory + hash_path)
    bytesio = BytesIO()
    df_write_csv(df, bytesio, has_header=False)
    path = f"{target_directory}{hash_path}/{append_column}_{str(df.row(0)[df.find_idx_by_name(append_column)])}.csv.gz"
    with FileLock(f'{path}.lock', timeout=60 * 60 * 2):
        if not exists(path):
            with gzip.open(path, 'wb') as f:
                f.write(column_header.encode('utf-8'))
        write_gzip(bytesio, path, 'ab')
    return path


def append_csv(df, d):
    hash_path = md5_hex(d)
    return append_csv_hash(df, d['target_directory'], d['append_column'], d['column_header'], hash_path)


def pipe(f, lst, dic, is_lock=True):
    yield from it_mp_f(f, [tuple([item, dic]) for item in lst], is_lock)


def zip_path(target_directory):
    path = target_directory + target_directory[target_directory.rfind('/', 0, -1) + 1:-1] + '.zip'
    return path


def make_zip(wildcard_source_path, target_path):
    with ZipFile(target_path, 'w') as f:
        for target_path in glob(wildcard_source_path):
            f.write(target_path)
    return target_path


def has_zip_cache(d):
    return exists(zip_cache_path(d))


def zip_cache(d):
    wildcard_source_path = f"{d['target_directory']}{md5_hex(d)}/*.csv.gz"
    if not glob(wildcard_source_path):
        return None
    return make_zip(wildcard_source_path, zip_cache_path(d))


def to_csv_gz(df, target_directory, url, lst_group_by_column):
    mkdir(target_directory)
    filename_tail = '_group_' + '-'.join(list(
        map(lambda column: str(df.row(0)[df.find_idx_by_name(column)]).replace('_', '-'), lst_group_by_column)))
    bytesio, filename = bytesio_filename(df, url, has_header=True)
    path = f'{target_directory}{filename}{filename_tail}.csv.gz'
    return write_gzip(bytesio, path)



from multiprocessing.dummy import Pool
from tqdm import tqdm
import cameo_claw.istarmap
from filelock import FileLock

cameo_claw.istarmap.ignore_no_use_warning()


def it_mp_f(f, tup_param, is_lock=True):
    if is_lock:
        # 2022-04-16 Bowen, 6 users query 3100 urls at the same time crashed, add FileLock to prevent it.
        with FileLock("it_mp_f.lock", timeout=60 * 60 * 2):
            yield from thread_pool_f(f, tup_param)
    else:
        yield from thread_pool_f(f, tup_param)


def thread_pool_f(f, tup_param):
    # 2022-04-10 bowen
    # fastapi + multiprocessing can crash, please use spawn to prevent it
    # https://miketarpey.medium.com/troubleshooting-usage-of-pythons-multiprocessing-module-in-a-fastapi-app-f1c368673686
    # it also fails!
    # finally, we use: from multiprocessing.dummy import Pool. Success!
    int_progress = 0
    with Pool(20) as p:
        for result in tqdm(p.istarmap(f, tup_param), total=len(tup_param)):
            int_progress += 1
            yield int_progress, result



from cameo_claw.client import Client
from cameo_claw.remote import square

client = Client()
lst = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
it = client.map(square, lst)
print(f'result:{next(it)}')
print(f'result:{next(it)}')
print(f'result:{next(it)}')



from setuptools import setup

setup(
    name='cameo_claw',
    version='1.2.4',
    description='Multiprocessing download, filter, streaming. ⚡️FAST⚡️ remove pandas',
    url='https://github.com/bohachu/cameo_claw',
    author='Bowen Chiu',
    author_email='bohachu@gmail.com',
    license='BSD 2-clause',
    packages=['cameo_claw'],
    install_requires=[
        'requests==2.23.0',
        'polars==0.13.24',
        'tqdm==4.63.1',
        'fastapi==0.75.0',
        'uvicorn==0.17.6',
        # 'pandas==1.4.1',
        'filelock==3.6.0',
        'glob2==0.7',
        'Jinja2==2.10.1',
        'pyarrow==7.0.0',
        'httpx==0.22.0',
        # 'asyncio==3.4.3'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
)
